/*******************************************************************
*****************************  realtime.c  *************************
********************************************************************
*
*  Better Windows NT real time frame generator
*
*  Developed by:  Lionell K. Griffith - March 31, 1998
*
*  email:  lgriffith@hughes.net
*
*     Procedure      Description  
*  1. Li2Double      Converts LONG_INTEGER value to double value
*  2. FastTimerInit  Initializes the Performance Timer
*  3. FastTimer      Returns performance timer time in seconds
*  4. FrameProc      Callback for frame timer
*  5. FrameSet       Sets and starts Frame timer
*  6. FrameKill      Terminates Frame timer
*  7. FrameTime      Returns Frame Start Time
*
*  NOTE:  winmm.lib must be in the linker library list
*
*******************************************************************/

#define STRICT

#include <windows.h>
#include <mmsystem.h>

#define BetterTimer TRUE

struct RealTimeStruct
  {
   LARGE_INTEGER TimerFrequency;  /* Performance timer frequency  */
   double Frequency;              /* Frequence converted to double*/
   MMRESULT hTimer;               /* Timer handle                 */
   double Period;                 /* Frame period                 */
   double FrameStart;             /* Start of frame time          */
   double Time;                   /* Frame time                   */
   double OldTime;                /* Old frame time               */
   double Error;                  /* Frame time error             */
   UINT Delay;                    /* Integer ms frame periods     */
   void (*pEntry)(void *);        /* Frame process entry pointer  */
   void *pData;                   /* Frame process data pointer   */
  };
typedef struct RealTimeStruct RealTimeType;

RealTimeType RealTime;            /* RealTime module data         */

/*******************************************************************
************************  1. Li2Double  ****************************
********************************************************************
*
*  Converts LONG_INTEGER value to double value
*
*  DoubleValue = Li2Double(LongInteger);
*
*  DoubleValue is the double value of the long integer
*  LongInteger is the LONG_INTEGER value to be converted
*
*******************************************************************/

#define C 4.294967296E9
#define Li2Double(x) (double)((x).HighPart)*C+(double)((x).LowPart)

/*******************************************************************
**************************  2. FastTimerInit  **********************
********************************************************************
*
*  Initializes the Performance Timer
*
*  FastTimerInit();
*
*******************************************************************/

void FastTimerInit(void)
  {
   QueryPerformanceFrequency(&RealTime.TimerFrequency);
   RealTime.Frequency = Li2Double(RealTime.TimerFrequency);
  }

/*******************************************************************
*****************************  3. FastTimer   **********************
********************************************************************
*
*  Returns performance timer time in seconds
*
*  Time = FastTimer();
*
*  Time is time in seconds
*
*  Timer resolution is 0.838 usec
*
*  The performance timer is a 64 bit counter.  It can count for
*  almost 700,000 years.  Hence, rollover is not a problem.
*
*******************************************************************/
 
double FastTimer(void)
  {
   LARGE_INTEGER CurrentTime;
   double Ticks;
   double Seconds;

   QueryPerformanceCounter(&CurrentTime);
   Ticks = Li2Double(CurrentTime);
   Seconds = Ticks/RealTime.Frequency;
   return Seconds;
  }

/*******************************************************************
****************************  4. FrameProc  ************************
********************************************************************
*
*   Callback for frame timer
*
*   FrameProc(uID,uMsg,dwUser,dw1,dw2);
*
*   uID Identifier of the timer event. This identifier was returned 
*        the timeSetEvent function when the timer event was set up. 
*
*   uMsg Reserved; do not use. 
*
*   dwUser  User instance data supplied to the dwUser parameter of 
*           timeSetEvent. 
*
*   dw1 and dw2  Reserved; do not use. 
*
*******************************************************************/

void CALLBACK FrameProc(UINT uID,UINT uMsg,DWORD dwUser, 
                        DWORD dw1,DWORD dw2)
  {

#if BetterTimer

   /* Disable hTimer */
   RealTime.hTimer = 0;

   /* Wait until period is over */
   RealTime.OldTime = RealTime.Time;
   do
     {
      RealTime.Time = FastTimer();
     }
   while (RealTime.Time-RealTime.OldTime < 
          RealTime.Period - RealTime.Error);

   RealTime.OldTime = RealTime.Time;

   /* Start next timer interval */
   RealTime.hTimer  = 
     timeSetEvent(RealTime.Delay,0,FrameProc,dwUser,TIME_ONESHOT);

   /* Capture Frame Time */
   RealTime.Time = FastTimer();

   /* Capture Frame Error */
   RealTime.Error = (RealTime.Time - RealTime.OldTime);

#else

   /* Capture Frame Time */
   RealTime.Time = FastTimer();

#endif

   /* Call timer callback procedure */
   if (RealTime.pEntry) (RealTime.pEntry)(RealTime.pData);
  }

/*******************************************************************
****************************  5. FrameSet   ************************
********************************************************************
* 
*   Sets and starts Frame timer
*
*   FrameSet(Period,pProcess,pData);
*
*   Period is frame period in seconds > 1 millisecond
*   pEntry is a pointer to the frame process entry point
*   pData is a pointer to the process data 
*
*******************************************************************/

void FrameSet(double Period,void *pEntry,void *pData)
  {
   RealTime.Period = Period;
   RealTime.pEntry = pEntry;
   RealTime.pData = pData;

#if BetterTimer

   RealTime.Delay =  (UINT)(Period*1000.0) - 1;
   RealTime.hTimer = 
     timeSetEvent(RealTime.Delay,0,FrameProc,0,TIME_ONESHOT);

#else

   RealTime.Delay =  (UINT)(Period*1000.0);
     RealTime.hTimer = 
       timeSetEvent(RealTime.Delay,0,FrameProc,0,TIME_PERIODIC);

#endif

   RealTime.Time = FastTimer();
   RealTime.Error = 0.0;
  }

/*******************************************************************
*************************  6. FrameKill  ***************************
********************************************************************
*   
*   Terminates Frame timer
*
*   FrameKill();
*
*******************************************************************/

void FrameKill(void)
  {
   if (RealTime.hTimer) 
     {
      timeKillEvent(RealTime.hTimer);
      RealTime.hTimer = 0;
     }
  }

/*******************************************************************
*************************  7. FrameTime  ***************************
********************************************************************
*   
*   Returns Frame Start Time
*
*   Time = FrameTime();
*
*   Time is the frame start time in seconds since powerup
*
*******************************************************************/

double FrameTime(void)
  {
   if (RealTime.hTimer) return RealTime.Time;
   else return 0.0;
  }

/*******************************************************************
************************  End of realtime.c  ***********************
*******************************************************************/

